#include "BaseService.h"


#ifdef _WIN32
    #pragma warning (push)
    #pragma warning (disable : 4355) // 'this' : used in base member initializer list
#endif


// Include autogenerated packet and coroutine definitions
#include "PacketSource.autogen.h"
#include "CoroutineSource.autogen.h"


#ifdef _WIN32
    #pragma warning (pop)
#endif


void ServiceList::RegisterService(ServiceId a_ServiceId, IService *a_pService)
{
    m_ServiceMap[a_ServiceId] = a_pService;
}


IService *ServiceList::GetService(ServiceId a_ServiceId) const
{
    ServiceMap::const_iterator iter = m_ServiceMap.find(a_ServiceId);

    if (iter == m_ServiceMap.end())
    {
        return NULL;
    }

    return iter->second;
}


BaseService::BaseService(ServiceId a_ServiceId, ServiceList &a_ServiceList)
:   m_ServiceId(a_ServiceId)
,   m_ServiceList(a_ServiceList)
{
    m_ServiceList.RegisterService(m_ServiceId, this);
}


ServiceId BaseService::GetServiceId() const
{
    return m_ServiceId;
}


bool BaseService::SendTo(ServiceId a_DstServiceId, IPacket &a_Packet)
{
    IService *pDestinationService = m_ServiceList.GetService(a_DstServiceId);

    if (pDestinationService == NULL)
    {
        WriteDebugString("BaseService: service lookup failed for serviceid %u, no route to service\n", (uint32_t)a_DstServiceId);
        return false;
    }

    PacketBuffer encodingBuffer;
    NetworkInputSerializer encoder(encodingBuffer, sizeof(encodingBuffer));

    // Encode header info and packet data for sending
    {
        PacketHeader header(GetServiceId(), a_Packet.GetPacketType());

        if (!encoder.Serialize(header, header.GetSerializableClassName()))
        {
            WriteDebugString("BaseService: serialization of packet header failed!\n");
            return false;
        }

        if (!encoder.Serialize(a_Packet, a_Packet.GetSerializableClassName()))
        {
            WriteDebugString("BaseService: serialization of packet %s failed!\n", a_Packet.GetSerializableClassName());
            return false;
        }
    }

    pDestinationService->InternalReceiveFrom(encoder.GetBuffer(), encoder.GetBufferSize());

    return true;
}


TimeoutId BaseService::RegisterItem(ITimeoutItem *a_pItem)
{
    return m_TimeoutQueue.RegisterItem(a_pItem);
}


ITimeoutItem *BaseService::RetrieveItem(TimeoutId a_TimeoutId)
{
    return m_TimeoutQueue.RetrieveItem(a_TimeoutId);
}


void BaseService::RemoveItem(TimeoutId a_TimeoutId)
{
    m_TimeoutQueue.RemoveItem(a_TimeoutId);
}


// Quick helper function to extract and resume a coroutine if one is bound to the packet
static bool ResumeCoroutine(IService *a_pService, IPacket &a_Packet)
{
    CoroutineMetadata coroutineRequestData;

    if (!a_Packet.GetMetadata().GetCoroutineData(coroutineRequestData))
    {
        return false;
    }

    // Verify that this coroutine actually belongs to us before proceeding
    if (coroutineRequestData.m_OwnerServiceId != a_pService->GetServiceId())
    {
        return false;
    }

    // There is in-fact a slight problem with this line
    // I've made the process of retrieving the coroutine implicitly refresh it's timeout counter
    // At this stage though, I don't actually know whether or not the coroutine is going to accept 
    // this packet and process it
    // In the case where we've timed out a previous operation, and then received a delayed response, 
    // this will incorrectly reset the timeout counter
    // I've decided the convenience here outweighs this particular issue, since the timeout queue 
    // makes no guarantees that timed out items will timeout immediately anyways
    ITimeoutItem *pCoroutine = a_pService->RetrieveItem(coroutineRequestData.m_TimeoutId);

    if (pCoroutine != NULL)
    {
        return pCoroutine->HandleAsyncCallback(a_Packet);
    }

    return false;
}


void BaseService::Update()
{
    MessageBuffer messagesToProcess;
    {
        // Swap out messages to process, can be guarded with a lock if multithreaded
        messagesToProcess.swap(m_MessageBuffer);
    }

    while (messagesToProcess.size() > 0)
    {
        QueuedMessage &message = messagesToProcess.back();

        if (!ResumeCoroutine(this, *(message.m_Message)))
        {
            RecvFrom(message.m_SendingServiceId, *(message.m_Message));
        }

        messagesToProcess.pop_back();
    }

    m_TimeoutQueue.UpdateTimeouts();
}


bool BaseService::InternalReceiveFrom(const uint8_t *a_pData, uint32_t a_iNumBytes)
{
    NetworkOutputSerializer decoder(a_pData, a_iNumBytes);

    PacketHeader header(ServiceId(0), EPacketType(0));

    if (!decoder.Serialize(header, header.GetSerializableClassName()))
    {
        WriteDebugString("BaseService: failed to serialize header from received packet!\n");
        return false;
    }

    std::shared_ptr<IPacket> pPacket(CreatePacket((EPacketType)header.m_PacketType));

    if (pPacket == NULL)
    {
        WriteDebugString("BaseService: unknown packet type %u encountered!\n", (uint32_t)header.m_PacketType);
        return false;
    }

    if (!decoder.Serialize(*pPacket, pPacket->GetSerializableClassName()))
    {
        WriteDebugString("BaseService: failed to serialize received packet %s!\n", pPacket->GetSerializableClassName());
        return false;
    }

    if (decoder.GetUnreadSize() > 0)
    {
        WriteDebugString("BaseService: failed serialization for received packet %s, unconsumed bytes remain in buffer!\n", pPacket->GetSerializableClassName());
        return false;
    }

    // Simply push-back the packet into our message buffer, we'll serialize and complete processing later
    {
        uint32_t iMessageIndex = (uint32_t)m_MessageBuffer.size();

        m_MessageBuffer.resize(iMessageIndex + 1);

        QueuedMessage &message = m_MessageBuffer[iMessageIndex];

        message.m_SendingServiceId = header.m_SendingServiceId;
        message.m_Message = pPacket;
    }

    return true;
}
