#ifndef ISERIALIZER_H
#define ISERIALIZER_H


#include <stdint.h>


class ISerializer;


/**
 * @class ISerializable
 * 
 * @brief Base class for serializable objects
 */
class ISerializable
{
public:

    /**
     * Virtual destructor
     */
    virtual ~ISerializable() {}

    /**
     * Gets the class name for the current object instance
     * 
     * @return string class name of the current object instance
     */
    virtual const char *GetSerializableClassName() const = 0;

    /**
     * Base serialize method for all serializable structures or classes to implement
     * 
     * @param a_Serializer ISerializable instance to use for serialization
     * @param a_Name       string name of the object
     * 
     * @return boolean true for success, false for serialization failure
     */
    virtual bool Serialize(ISerializer &a_Serializer, const char *a_Name) = 0;

};


/**
 * @class ISerializable
 * 
 * @brief Interface class for all Serializers to derive from
 */
class ISerializer
{
public:

    /**
     * Virtual destructor
     */
    virtual ~ISerializer() {}

    /**
     * Returns true if the serializer is writing values to the objects it visits
     * This interface is required for efficiently serializing most dynamic data-structures
     * 
     * @return boolean true if the serializer is writing to objects that it visits
     */
    virtual bool IsWritingToObject() const = 0;

    /**
     * Serialize a boolean
     * 
     * @param a_Value boolean input value to serialize
     * @param a_Name  string name of the value being serialized
     * 
     * @return boolean true for success, false for serialization failure
     */
    virtual bool Serialize(bool &a_Value, const char *a_Name) = 0;

    /**
     * Serialize a character
     * 
     * @param a_Value character input value to serialize
     * @param a_Name  string name of the value being serialized
     * 
     * @return boolean true for success, false for failure
     */
    virtual bool Serialize(char &a_Value, const char *a_Name) = 0;

    /**
     * Serialize a signed byte
     * 
     * @param a_Value signed byte input value to serialize
     * @param a_Name  string name of the value being serialized
     * 
     * @return boolean true for success, false for failure
     */
    virtual bool Serialize(int8_t &a_Value, const char *a_Name) = 0;

    /**
     * Serialize a signed short
     * 
     * @param a_Value signed short input value to serialize
     * @param a_Name  string name of the value being serialized
     * 
     * @return boolean true for success, false for failure
     */
    virtual bool Serialize(int16_t &a_Value, const char *a_Name) = 0;

    /**
     * Serialize a signed integer
     * 
     * @param a_Value signed integer input value to serialize
     * @param a_Name  string name of the value being serialized
     * 
     * @return boolean true for success, false for failure
     */
    virtual bool Serialize(int32_t &a_Value, const char *a_Name) = 0;

    /**
     * Serialize a signed 64-bit integer
     * 
     * @param a_Value signed 64-bit integer input value to serialize
     * @param a_Name  string name of the value being serialized
     * 
     * @return boolean true for success, false for failure
     */
    virtual bool Serialize(int64_t &a_Value, const char *a_Name) = 0;

    /**
     * Serialize an unsigned byte
     * 
     * @param a_Value unsigned byte input value to serialize
     * @param a_Name  string name of the value being serialized
     * 
     * @return boolean true for success, false for failure
     */
    virtual bool Serialize(uint8_t &a_Value, const char *a_Name) = 0;

    /**
     * Serialize an unsigned short
     * 
     * @param a_Value signed integer short value to serialize
     * @param a_Name  string name of the value being serialized
     * 
     * @return boolean true for success, false for failure
     */
    virtual bool Serialize(uint16_t &a_Value, const char *a_Name) = 0;

    /**
     * Serialize an unsigned integer
     * 
     * @param a_Value signed integer input value to serialize
     * @param a_Name  string name of the value being serialized
     * 
     * @return boolean true for success, false for failure
     */
    virtual bool Serialize(uint32_t &a_Value, const char *a_Name) = 0;

    /**
     * Serialize an unsigned 64-bit integer
     * 
     * @param a_Value signed 64-bit integer input value to serialize
     * @param a_Name  string name of the value being serialized
     * 
     * @return boolean true for success, false for failure
     */
    virtual bool Serialize(uint64_t &a_Value, const char *a_Name) = 0;

    /**
     * Serialize a serializable object
     * 
     * @param a_Value object instance to serialize
     * @param a_Name  string name of the object
     * 
     * @return boolean true for success, false for serialization failure
     */
    virtual bool Serialize(ISerializable &a_Value, const char *a_Name) = 0;

};


#endif
