#ifndef ISERVICE_H
#define ISERVICE_H


#include "StrongTypedef.h"


STRONG_TYPEDEF(uint32_t, ServiceId);
STRONG_TYPEDEF(uint32_t, TimeoutId);


// Forward declare
class IPacket;
class ITimeoutItem;


/**
 * @class IService
 * 
 * @brief Base class emulating a message-based service in a distributed architecture
 */
class IService
{
public:

    /**
     * Virtual destructor
     */
    virtual ~IService() {}

    /**
     * Returns a unique service identifer
     * Could be a hash of address, process Id, etc..
     * 
     * @return unique service id
     */
    virtual ServiceId GetServiceId() const = 0;

    /**
     * Sends a message to the give service Id
     * 
     * @param a_DstServiceId identifier of the service to send to
     * @param a_Packet       packet to send
     * 
     * @return boolean true on success
     */
    virtual bool SendTo(ServiceId a_DstServiceId, IPacket &a_Packet) = 0;

    /**
     * Receive a message from a service
     * 
     * @param a_SrcServiceId identifer of the service sending the message
     * @param a_Packet       received packet
     * 
     * @return boolean true on success
     */
    virtual bool RecvFrom(ServiceId a_SrcServiceId, IPacket &a_Packet) = 0;

    /**
     * Registers a new item with the services timeout mechanism
     * 
     * @param a_Item instance of the item to register
     * 
     * @return boolean true if registration was successful
     */
    virtual TimeoutId RegisterItem(ITimeoutItem *a_pItem) = 0;

    /**
     * Returns the provided timeout item if it exists, also refreshes the timeout value
     * 
     * @param a_TimeoutId the identifier of the item to fetch
     * 
     * @return pointer to the timeout item if it exists
     */
    virtual ITimeoutItem *RetrieveItem(TimeoutId a_TimeoutId) = 0;

    /**
     * Removes an item from the TimeoutQueue
     * 
     * @param a_TimeoutId the identifier of the item to remove
     */
    virtual void RemoveItem(TimeoutId a_TimeoutId) = 0;

    /**
     * Updates the service
     */
    virtual void Update() = 0;

    /**
     * Internal packet receive mechanism, network layer could use this method
     * 
     * @param a_pData     received data
     * @param a_iNumBytes size of a_pData in bytes
     * 
     * @return boolean true on success
     */
    virtual bool InternalReceiveFrom(const uint8_t *a_pData, uint32_t a_iNumBytes) = 0;

};


#endif
