#include "BaseService.h"
#include <assert.h>
#include <stdio.h>


#ifdef _WIN32
    #define WIN32_LEAN_AND_MEAN
    #include <windows.h>
#endif


// Simple test of serialization
// Construct a packet, set a bunch of members, encode/decode to new instance, and compare the results
bool TestSerialization()
{
    Foo testFoo1;

    testFoo1.GetMetadata().SetRequestService(ServiceId(1));
    testFoo1.SetBoolean(true);
    testFoo1.SetCharacter('c');
    testFoo1.SetSignedByte(-1);
    testFoo1.SetSignedShort(-1);
    testFoo1.SetSignedInt(-1);
    testFoo1.SetSignedLong(-1);
    testFoo1.SetUnsignedByte(~0);
    testFoo1.SetUnsignedShort(~0);
    testFoo1.SetUnsignedInt(~0);
    testFoo1.SetUnsignedLong(~0);

    PacketBuffer SerializationBuffer;

    NetworkInputSerializer input(SerializationBuffer, sizeof(SerializationBuffer));

    if (!input.Serialize(testFoo1, testFoo1.GetSerializableClassName()))
    {
        return false;
    }

    NetworkOutputSerializer output(input.GetBuffer(), input.GetBufferSize());

    Foo testFoo2;

    if (!output.Serialize(testFoo2, testFoo2.GetSerializableClassName()))
    {
        return false;
    }

    return (testFoo1 == testFoo2);
}


// Fake a client
class ExampleClientService
:   public BaseService
{
public:
    ExampleClientService(ServiceId a_ServiceId, ServiceList &a_ServiceList) : BaseService(a_ServiceId, a_ServiceList), m_bDone(false), m_iResult(0) {}
    
    bool RecvFrom(ServiceId a_SrcServiceId, IPacket &a_Packet)
    {
        WriteDebugString("ExampleClientService, received %s\n", a_Packet.GetSerializableClassName());

        // Only care about handling the two expected responses
        const SumTenRequestSuccess *pSuccess = SafeCastPacket<SumTenRequestSuccess>(&a_Packet);
        if (pSuccess != NULL)
        {
            m_iResult = pSuccess->GetResult();
            WriteDebugString("ExampleClientService: done with result %d!\n", pSuccess->GetResult());
            m_bDone = true;
            return true;
        }

        const SumTenRequestError *pError = SafeCastPacket<SumTenRequestError>(&a_Packet);
        if (pError != NULL)
        {
            WriteDebugString("ExampleClientService: failed!\n");
            m_bDone = true;
            return true;
        }

        return false;
    }

    bool     m_bDone;
    uint32_t m_iResult;
};


// Fake a distributed endpoint, this service is responsible for calculating sums of inputs!
class ExampleSumService
:   public BaseService
{
public:
    ExampleSumService(ServiceId a_ServiceId, ServiceList &a_ServiceList) : BaseService(a_ServiceId, a_ServiceList) {}
    
    bool RecvFrom(ServiceId a_SrcServiceId, IPacket &a_Packet)
    {
        WriteDebugString("ExampleSumService, received %s\n", a_Packet.GetSerializableClassName());

        const SumRequest *pRequest = SafeCastPacket<SumRequest>(&a_Packet);
        if (pRequest != NULL)
        {
            // Create a response packet
            SumRequestSuccess result(pRequest->GetMetadata());

            // Do the sum
            result.SetResult(pRequest->GetBaseValue() + pRequest->GetAmountToAdd());

            // Send the response
            WriteDebugString("ExampleSumService: ok! %d + %d = %d\n", pRequest->GetBaseValue(), pRequest->GetAmountToAdd(), result.GetResult());

            // Drops the request by never sending a response, this forces the coroutine to timeout its request
            // Also causes an assert, since the coroutine output is verified for correctness below
            bool bForceFail = false;
            if (!bForceFail)
            {
                SendTo(a_SrcServiceId, result);
            }
            return true;
        }

        return false;
    }
};


// Fake a coroutine driven service
// Calculates the result of the sequence 0 + 1 + 2 + 3 + ... + 9
// All the actual summations are offloaded to the SumService, defined above
class ExampleCoroutineService
:   public BaseService
{
public:
    ExampleCoroutineService(ServiceId a_ServiceId, ServiceList &a_ServiceList) : BaseService(a_ServiceId, a_ServiceList) {}
    
    bool RecvFrom(ServiceId a_SrcServiceId, IPacket &a_Packet)
    {
        WriteDebugString("ExampleCoroutineService, received %s\n", a_Packet.GetSerializableClassName());

        SumTenRequest *pRequest = SafeCastPacket<SumTenRequest>(&a_Packet);
        if (pRequest != NULL)
        {
            WriteDebugString("ExampleCoroutineService: start!\n");

            // Start a coroutine to handle this...
            SumTenDistributed::Coroutine *pCoroutine = new SumTenDistributed::Coroutine(this, a_SrcServiceId, *pRequest);
            pCoroutine->Run();
            return true;
        }

        return false;
    }
};


// Simple test of the coroutine system, create a client, sum, and coroutine service, invoke everything and grab the result when done
bool TestCoroutines()
{
    ServiceList serviceList;

    ExampleClientService       clientService(ServiceId(1), serviceList);
    ExampleSumService             sumService(ServiceId(2), serviceList);
    ExampleCoroutineService coroutineService(ServiceId(3), serviceList);

    // Send a request to the coroutine service
    SumTenRequest request;
    clientService.SendTo(ServiceId(3), request);

    while (!clientService.m_bDone)
    {
        // Update our fake services
           clientService.Update();
              sumService.Update();
        coroutineService.Update();
    }

    if (clientService.m_iResult != 45) // 0 + 1 + 2 + 3 + 4 + 5 + 6 + 7 + 8 + 9
    {
        return false;
    }

    return true;
}


int main(int argc, char **argv)
{
    if (!TestSerialization())
    {
        WriteDebugString("Packet serialization failed\n");
        assert(0);
        return 1;
    }

    if (!TestCoroutines())
    {
        WriteDebugString("Coroutine failed\n");
        assert(0);
        return 1;
    }

    return 0;
}
