#include "NetworkInputSerializer.h"
#include <memory>


#ifdef _WIN32
    #include <Winsock2.h>
#elif defined __APPLE_CC__
    #include <Machine/endian.h>
#else
    #include <endian.h>
    #include <string.h>
    #include <arpa/inet.h>
#endif


// It works, but many don't like this
// Easy enough to add a run-time check instead if you really want, but my goal here is simply minimal code
#if __BYTE_ORDER__ == __LITTLE_ENDIAN
    #define htonll(X) (((uint64_t)htonl((uint32_t)(X))) << 32 | (uint64_t)htonl((uint32_t)(X >> 32)))
#else
    #define htonll(X) X
#endif


NetworkInputSerializer::NetworkInputSerializer(uint8_t *a_pBuffer, uint32_t a_iBufferCapacity)
:   m_bSerializationFailed(false) 
,   m_iBufferSize(0)
,   m_pBuffer(a_pBuffer)
,   m_iBufferCapacity(a_iBufferCapacity)
{
    ;
}


bool NetworkInputSerializer::CopyToBuffer(const uint8_t *a_pData, uint32_t a_iDataSize)
{
    return SerializeBytes(a_pData, a_iDataSize);
}


const uint8_t *NetworkInputSerializer::GetBuffer() const
{
    return m_pBuffer;
}


uint32_t NetworkInputSerializer::GetBufferSize() const
{
    return m_iBufferSize;
}


bool NetworkInputSerializer::IsWritingToObject() const
{
    return false;
}


bool NetworkInputSerializer::Serialize(bool &a_Value, const char *a_Name)
{
    uint8_t iValue = (a_Value) ? 1 : 0;
    return Serialize(iValue, a_Name);
}


bool NetworkInputSerializer::Serialize(char &a_Value, const char *a_Name)
{
    return SerializeBytes((const uint8_t *)&a_Value, sizeof(char));
}


bool NetworkInputSerializer::Serialize(int8_t &a_Value, const char *a_Name)
{
    return SerializeBytes((const uint8_t *)&a_Value, sizeof(int8_t));
}


bool NetworkInputSerializer::Serialize(int16_t &a_Value, const char *a_Name)
{
    int16_t iNetworkOrder = htons(a_Value);
    return SerializeBytes((const uint8_t *)&iNetworkOrder, sizeof(int16_t));
}


bool NetworkInputSerializer::Serialize(int32_t &a_Value, const char *a_Name)
{
    int32_t iNetworkOrder = htonl(a_Value);
    return SerializeBytes((const uint8_t *)&iNetworkOrder, sizeof(int32_t));
}


bool NetworkInputSerializer::Serialize(int64_t &a_Value, const char *a_Name)
{
    int64_t iNetworkOrder = htonll(a_Value);
    return SerializeBytes((const uint8_t *)&iNetworkOrder, sizeof(int64_t));
}


bool NetworkInputSerializer::Serialize(uint8_t &a_Value, const char *a_Name)
{
    return SerializeBytes((const uint8_t *)&a_Value, sizeof(uint8_t));
}


bool NetworkInputSerializer::Serialize(uint16_t &a_Value, const char *a_Name)
{
    uint16_t iNetworkOrder = htons(a_Value);
    return SerializeBytes((const uint8_t *)&iNetworkOrder, sizeof(uint16_t));
}


bool NetworkInputSerializer::Serialize(uint32_t &a_Value, const char *a_Name)
{
    uint32_t iNetworkOrder = htonl(a_Value);
    return SerializeBytes((const uint8_t *)&iNetworkOrder, sizeof(uint32_t));
}


bool NetworkInputSerializer::Serialize(uint64_t &a_Value, const char *a_Name)
{
    uint64_t iNetworkOrder = htonll(a_Value);
    return SerializeBytes((const uint8_t *)&iNetworkOrder, sizeof(uint64_t));
}


bool NetworkInputSerializer::Serialize(ISerializable &a_Value, const char *a_Name)
{
    return a_Value.Serialize(*this, a_Name);
}


bool NetworkInputSerializer::SerializeBytes(const uint8_t *a_pData, uint32_t a_iCount)
{
    const uint32_t iCurrSize = m_iBufferSize;
    const uint32_t iNextSize = m_iBufferSize + a_iCount;

    if (m_bSerializationFailed || (iNextSize >= m_iBufferCapacity))
    {
        // Keep the failed boolean as a paranoid check
        // If everyone honours the serialize boolean return values and terminates on fail, then this check is redundant
        // Otherwise we might fail to serialize a large value, and then succeed a smaller value
        m_bSerializationFailed = true;
        return false;
    }

    uint8_t *pWriteBuffer = (uint8_t *)(m_pBuffer + iCurrSize);
    memcpy(pWriteBuffer, a_pData, a_iCount);

    m_iBufferSize += a_iCount;

    return true;
}
