#ifndef STRONG_TYPEDEF_H
#define STRONG_TYPEDEF_H


#include "ISerializer.h"


// Icky function pointer, so that I can pretend to pass a const char * as a template parameter
typedef const char *(*NameFunc)(void);


// Use this macro in place of typdef
// Example: typedef float foo;
//            becomes 
//            STRONG_TYPEDEF(float, foo);
#define STRONG_TYPEDEF(TYPE, ALIAS) \
    inline const char *GetName##ALIAS() { return #ALIAS; } \
    typedef StrongTypedef<TYPE, GetName##ALIAS> ALIAS


/**
 * @class StrongTypedef
 * 
 * @brief Helper class, you can use boost strong typedef instead if you want
 * 
 * This is a quick-and-dirty implementation of a strongly typed typedef, so I don't need to include all of boost
 * It also has the benefit of adhering to the serialization interface, so I can pull some nice tricks with the factory
 */
template <typename BASE_TYPE, NameFunc NAME_FUNC>
class StrongTypedef : public ISerializable
{
public:

    // Just in-case you require compile-time access to the inner type, this typedef provides you with ALIAS::Type
    typedef BASE_TYPE Type;

    /**
     * Default constructor
     */
    StrongTypedef();

    /**
     * Copy construct from same type
     * 
     * @param a_RHS instance to construct from
     */
    StrongTypedef(const StrongTypedef<BASE_TYPE, NAME_FUNC> &a_RHS);

    /**
     * Copy construct from underlying base type
     * 
     * @param a_RHS base type value to construct from
     */
    explicit StrongTypedef(const BASE_TYPE &a_RHS);

    /**
     * Assignment from same type
     * 
     * @param a_RHS instance to assign from
     */
    StrongTypedef<BASE_TYPE, NAME_FUNC> &operator = (const StrongTypedef &a_RHS);

    /**
     * Assignment from underlying base type
     * 
     * @param a_RHS base type value to assign from
     */
    StrongTypedef<BASE_TYPE, NAME_FUNC> &operator = (const BASE_TYPE &a_RHS);

    /**
     * Const base type operator
     * 
     * @return value in const base type form
     */
    operator const BASE_TYPE& () const;

    /**
     * Non-const base type operator
     * 
     * @return value in base type form
     */
    operator BASE_TYPE& ();

    /**
     * Equality operator
     * 
     * @param a_RHS base type value to compare against
     * 
     * @return boolean true if this == a_RHS
     */
    bool operator == (const StrongTypedef<BASE_TYPE, NAME_FUNC> &a_RHS) const;

    /**
     * Inequality operator
     * 
     * @param a_RHS base type value to compare against
     * 
     * @return boolean true if this != a_RHS
     */
    bool operator != (const StrongTypedef<BASE_TYPE, NAME_FUNC> &a_RHS) const;

    /**
     * Strictly less than operator
     * 
     * @param a_RHS base type value to compare against
     * 
     * @return boolean true if this < a_RHS
     */
    bool operator < (const StrongTypedef<BASE_TYPE, NAME_FUNC> &a_RHS) const;

    /**
     * Less than equal to operator
     * 
     * @param a_RHS base type value to compare against
     * 
     * @return boolean true if this <= a_RHS
     */
    bool operator <= (const StrongTypedef<BASE_TYPE, NAME_FUNC> &a_RHS) const;

    /**
     * Strictly greater than operator
     * 
     * @param a_RHS base type value to compare against
     * 
     * @return boolean true if this > a_RHS
     */
    bool operator > (const StrongTypedef<BASE_TYPE, NAME_FUNC> &a_RHS) const;

    /**
     * Greater than equal to operator
     * 
     * @param a_RHS base type value to compare against
     * 
     * @return boolean true if this >= a_RHS
     */
    bool operator >= (const StrongTypedef<BASE_TYPE, NAME_FUNC> &a_RHS) const;

    /**
     * ISerializable interface
     * 
     * @return string class name of the current object instance
     */
    virtual const char *GetSerializableClassName() const;

    /**
     * ISerializable interface
     * 
     * @param a_Serializer ISerializable instance to use for serialization
     * @param a_Name       string name of the object
     * 
     * @return boolean true for success, false for serialization failure
     */
    virtual bool Serialize(ISerializer &a_Serializer, const char *a_Name);

private:

    BASE_TYPE m_Type;

};


#include "StrongTypedef.inl"


#endif
