#ifndef TIMEOUT_QUEUE_H
#define TIMEOUT_QUEUE_H


#include "CommonUtils.h"
#include "IService.h"
#include <memory>
#include <queue>
#include <map>


// Forward declare
class IPacket;


enum ETimeoutResult
{
    ETimeoutResult_Refresh, 
    ETimeoutResult_Delete
};


/**
 * @class ITimeoutItem
 * 
 * @brief interface class for objects that require timeout behaviour
 */
class ITimeoutItem
{
public:

    /**
     * Virtual destructor
     */
    virtual ~ITimeoutItem() {}

    /**
     * Asynchronous callback handler, yeah this a minor abuse of this interface
     * I've actually combined two abstractions here to just to reduce the amount of code required
     * In MWO, we have an extra layer of indirection between 'timeout' objects, and 'requestdata' objects with asynchronous behaviour
     * 
     * @param a_Result the asynchronous result to callback into the timeout instance with
     */
    virtual bool HandleAsyncCallback(IPacket &a_Result) = 0;

    /**
     * Timeout function, gets called when an instance exceeds it's timeout
     * This interface actually belongs here
     * 
     * @return ETimeoutResult_Refresh if you want to be re-queued for timeout
     *         ETimeoutResult_Delete  if you want the timeout queue to simply delete the instance
     */
    virtual ETimeoutResult OnTimeout() = 0;

    /**
     * Returns the timeout time in milliseconds to use for this object instance
     * 
     * @return the timeout time in milliseconds to use for this object instance
     */
    virtual uint32_t GetTimeoutMS() const = 0;
};


/**
 * @struct CoroutineMetadata
 * 
 * @brief container for all the indentifiers required to safely handle responses to coroutines
 */
struct CoroutineMetadata
:   public ISerializable
{
    ServiceId m_OwnerServiceId;
    TimeoutId m_TimeoutId;
    uint16_t  m_AsyncRequestId;

    virtual const char *GetSerializableClassName() const
    {
        return "CoroutineMetadata";
    }

    virtual bool Serialize(ISerializer &a_Serializer, const char *a_Name)
    {
        return a_Serializer.Serialize(m_OwnerServiceId, "OwnerServiceId") && 
               a_Serializer.Serialize(m_TimeoutId,      "TimeoutId") && 
               a_Serializer.Serialize(m_AsyncRequestId, "AsyncRequestId");
    }

    bool operator ==(const CoroutineMetadata &a_RHS) const
    {
        return m_OwnerServiceId == a_RHS.m_OwnerServiceId &&
               m_TimeoutId      == a_RHS.m_TimeoutId      &&
               m_AsyncRequestId == a_RHS.m_AsyncRequestId;
    }
};


/**
 * @class TimeoutQueue
 * 
 * @brief class for managing timeout items
 */
class TimeoutQueue
{
public:

    /**
     * Default constructor
     */
    TimeoutQueue();

    /**
     * Destructor
     */
    ~TimeoutQueue();

    /**
     * Registers a new item with the TimeoutQueue
     * 
     * @param a_Item instance of the item to register
     * 
     * @return boolean true if registration was successful
     * 
     *  **** TimeoutQueue ASSUMES OWNERSHIP ****
     */
    TimeoutId RegisterItem(ITimeoutItem *a_pItem);

    /**
     * Returns the provided timeout item if it exists, also refreshes the timeout value
     * 
     * @param a_TimeoutId the identifier of the item to fetch
     * 
     * @return pointer to the timeout item if it exists
     */
    ITimeoutItem *RetrieveItem(TimeoutId a_TimeoutId);

    /**
     * Removes an item from the TimeoutQueue
     * 
     * @param a_TimeoutId the identifier of the item to remove
     */
    void RemoveItem(TimeoutId a_TimeoutId);

    /**
     * Updates timeouts for all items, invokes timeout handlers if required
     */
    void UpdateTimeouts();

private:

    struct TimeoutQueueItem
    {
        TimeoutQueueItem(TimeoutId a_TimeoutId, uint32_t a_TimeoutTimeMS);

        bool operator < (const TimeoutQueueItem &a_RHS) const;

        TimeoutId m_TimeoutId;
        uint32_t  m_TimeoutTimeMS;
    };

    struct TimeoutMapItem
    {
        TimeoutMapItem();

        TimeoutMapItem(ITimeoutItem *a_pItem, uint32_t a_TimeoutTimeMS);

        std::shared_ptr<ITimeoutItem> m_pItem;
        uint32_t m_TimeoutTimeMS;
    };

    typedef std::map<TimeoutId, TimeoutMapItem  > TimeoutItemMap;
    typedef std::priority_queue<TimeoutQueueItem> TimeoutItemQueue;

    TimeoutId        m_NextTimeoutId;

    TimeoutItemMap   m_TimeoutItemMap;

    TimeoutItemQueue m_TimeoutItemQueue;

};


#include "TimeoutQueue.inl"


#endif
